local L = AceLibrary("AceLocale-2.2"):new("BuffMe")
local BS = AceLibrary("Babble-Spell-2.2")
local aura = AceLibrary("SpecialEvents-Aura-2.0")
local compost = AceLibrary("Compost-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local gratuity = AceLibrary("Gratuity-2.0")

BuffMe = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")

local function SortTable(tbl, f)
	local array = compost:Acquire()

	for num in pairs(tbl) do
		table.insert(array, num)
	end
	table.sort(array, f)
	local i = 0
	local iter = function ()
		i = i + 1
		if (array[i] == nil) then
			return nil
		else
			return array[i], tbl[array[i]]
		end
	end
	return iter
end

function BuffMe:OnInitialize()
	local options = {
		type = 'group',
		args = {
			hide = {
				type = 'toggle',
				name = L["Hide"],
				desc = L["Do not show buffs you currently have"],
				get = function() return self.db.profile.hide end,
				set = function(v)
					self.db.profile.hide = v
					self:UpdateBuffs()
				end,
			},
			autohide = {
				type = 'toggle',
				name = L["Auto-hide"],
				desc = L["Automatically hide window when ALL buffs are active"],
				get = function() return self.db.profile.autohide end,
				set = function(v)
					self.db.profile.autohide = v
					self:UpdateBuffs()
				end,
				disabled = function() return not self.db.profile.hide end,
			},
			lock = {
				type = 'toggle',
				name = L["Lock"],
				desc = L["Prevents the window from being moved"],
				get = function() return self.db.profile.lock end,
				set = function(v) self.db.profile.lock = v end,
			},
			clamped = {
				type = 'toggle',
				name = L["Clamped"],
				desc = L["Stop the window from being dragged off-screen"],
				get = function() return self.db.profile.clamped end,
				set = function(v) self.db.profile.clamped = v end,
			},
			alpha = {
				type = 'range',
				name = L["Alpha"],
				desc = L["Set the alpha of the frame"],
				isPercent = true,
				get = function() return self.db.profile.alpha end,
				set = function(v)
					self.db.profile.alpha = v
					self:ApplySettings()
				end,
			},
			scale = {
				type = 'range',
				name = L["Scale"],
				desc = L["Size the frame depending on your needs"],
				min = 0.5,
				max = 1.5,
				isPercent = true,
				get = function() return self.db.profile.scale end,
				set = function(v)
					self.db.profile.scale = v
					self:ApplySettings()
				end,
			},
			width = {
				type = 'range',
				name = L["Width"],
				desc = L["Adjust the frame's width"],
				min = 100,
				max = 400,
				get = function() return self.db.profile.width end,
				set = function(v)
					self.db.profile.width = v
					self:ApplySettings()
				end,
			},
			reset = {
				type = 'execute',
				name = L["Reset"],
				desc = L["Restore all settings to default"],
				func = function()
					self:ResetDB("profile")
					self:ApplySettings()
					self:UpdateBuffs()
					self:Print("All settings have now been reset to default")
				end,
			},
			border = {
				type = 'group',
				name = L["Border"],
				desc = L["Options for setting the border"],
				args = {
					alpha = {
						type = 'range',
						name = L["Alpha"],
						desc = L["Set the alpha of the border"],
						isPercent = true,
						get = function() return self.db.profile.borderAlpha end,
						set = function(v)
							self.db.profile.borderAlpha = v
							self:ApplySettings()
						end,
					},
					color = {
						type = 'color',
						name = L["Color"],
						desc = L["Set the color of the border"],
						get = function() return self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b end,
						set = function(r, g, b)
							self.db.profile.borderColor = { r = r, g = g, b = b }
							self:ApplySettings()
						end,
					},
					texture = {
				 		type = 'text',
						name = L["Texture"],
						desc = L["Border texture"],
						validate = {"none", "default", "tooltip"},
						get = function() return self.db.profile.texture end,
						set = function(v)
							self.db.profile.texture = v
							self:ApplySettings()
						end,
					},
				},
			},			
			backdrop = {
				type = 'group',
				name = L["Backdrop"],
				desc = L["Various options for the frame's backdrop"],
				args = {
					alpha = {
						type = 'range',
						name = L["Alpha"],
						desc = L["Set the alpha of the backdrop"],
						isPercent = true,
						get = function() return self.db.profile.backdropAlpha end,
						set = function(v)
							self.db.profile.backdropAlpha = v
							self:ApplySettings()
						end,
					},
					color = {
						type = 'color',
						name = L["Color"],
						desc = L["Set the color of the backdrop"],
						get = function() return self.db.profile.backdropColor.r, self.db.profile.backdropColor.g, self.db.profile.backdropColor.b end,
						set = function(r, g, b)
							self.db.profile.backdropColor = { r = r, g = g, b = b }
							self:ApplySettings()
						end,
					},
				},
			},			
		},
	}

	local defaults = {
		hide = true,
		autohide = true,
		lock = true,
		alpha = 0.8,
		clamped = true,
		scale = 1,
		borderAlpha = 1,
		borderColor = compost:AcquireHash(r, 0.5, g, 0.5, b, 0.5),
		backdropAlpha = 1,
		backdropColor = compost:AcquireHash(r, 0, g, 0, b, 0),
		texture = "none",
		posX = 1000,
		posY = 200,
		width = 224,
	}
	
	local defaults_char = {
		buffs = compost:Acquire(),
		weaponBuffs = compost:Acquire(),
	}

	options.args.config = {
		type = 'execute',
		name = L["GUI Config"],
		desc = L["Graphical options menu"],
		guiHidden = true,
		func =  function()
			dewdrop:Open(UIParent, 'children', function() dewdrop:FeedAceOptionsTable(options) end, 'cursorX', true, 'cursorY', true)
		end,
	}

	--self.frame = BuffMeFrame
	self.numBuffs = 0
	self.numBuffsShown = 0
	self.tooltip = BuffMeTooltip
	self.textures = {
		["none"]	= nil,
		["default"]	= "Interface\\DialogFrame\\UI-DialogBox-Border",
		["tooltip"]	= "Interface\\Tooltips\\UI-Tooltip-Border",
	}
	self.activeBuffs = compost:Acquire()
	
	self:RegisterDB("BuffMeDB", "BuffMePerCharDB")
	self:RegisterDefaults("profile", defaults)
	self:RegisterDefaults("char", defaults_char)
	self:RegisterChatCommand({L["SLASHCMD_LONG"], L["SLASHCMD_SHORT"]}, options)

	self:InitFrame()

	if (not dewdrop:IsRegistered(BuffMeFrame)) then
		dewdrop:Register(BuffMeFrame,
			'children', function(level, value) self:MenuSettings(level, value) end,
			'point', "TOPRIGHT",
			'relativePoint', "BOTTOMLEFT",
			'cursorX', true,
			'cursorY', true)
	end
end

function BuffMe:OnEnable()
	self:RegisterEvent("SpecialEvents_PlayerBuffGained", "UpdateBuffs")
	self:RegisterEvent("SpecialEvents_PlayerBuffLost", "UpdateBuffs")

	self:LoadBuffs()
	self:UpdateBuffs()
	BuffMeFrame:Show()
end

function BuffMe:OnDisable()
	BuffMeFrame:Hide()
end

function BuffMe:InitFrame()
	local opt = self.db.profile
	local f = CreateFrame("Button", "BuffMeFrame", UIParent)

	f:SetWidth(opt.width)
	f:SetHeight(100)

	f:SetClampedToScreen(opt.clamped)
	f:SetMovable(true)
	f:RegisterForDrag("LeftButton", "RightButton")
	f:SetScript("OnMouseDown", function() if (not opt.lock) and (arg1 == "LeftButton") then f:StartMoving() else self:UpdateBuffs() end end)
	f:SetScript("OnMouseUp", function() f:StopMovingOrSizing() self:ApplySettings() end)

	f:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile = self.textures[opt.texture], tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})

	f:SetAlpha(opt.alpha)
	f:SetBackdropBorderColor(opt.borderColor.r, opt.borderColor.g, opt.borderColor.b, opt.borderAlpha)
	f:SetBackdropColor(opt.backdropColor.r, opt.backdropColor.g, opt.backdropColor.b, opt.backdropAlpha)

	f:SetScale(opt.scale)

	f:ClearAllPoints()
	f:SetPoint("TOPLEFT", UIParent, "TOPLEFT", opt.posX, opt.posY)
	
	f:Hide()
end

function BuffMe:ApplySettings()
	local f = getglobal("BuffMeFrame")
	local opt = self.db.profile

	f:SetWidth(opt.width)
	f:SetScale(opt.scale)

	f:SetClampedToScreen(opt.clamped)

	fbd = f:GetBackdrop(fbd)
	fbd.edgeFile = self.textures[opt.texture]

	f:SetBackdrop(fbd)
	
	f:SetAlpha(opt.alpha)
	f:SetBackdropBorderColor(opt.borderColor.r, opt.borderColor.g, opt.borderColor.b, opt.borderAlpha)
	f:SetBackdropColor(opt.backdropColor.r, opt.backdropColor.g, opt.backdropColor.b, opt.backdropAlpha)
	
	_,_,_, posX, posY = f:GetPoint()
	opt.posX = posX
	opt.posY = posY
	
	for i = 1, self.numBuffs do
		local buff = getglobal("BuffMeFrameBuff"..i.."Text")
		buff:SetWidth(self.db.profile.width - 24)
	end
end

function BuffMe:MenuSettings(level, value)
	if (level == 1) then
		dewdrop:AddLine('text', "BuffMe! Options", 'isTitle', true)
		dewdrop:AddLine()

		for class in SortTable(self.buffs) do
			dewdrop:AddLine('text', class, 'hasArrow', true, 'value', class.."Buffs", 'tooltipTitle', "")
		end

		dewdrop:AddLine()

		for category in pairs(self.others) do
			dewdrop:AddLine('text', category, 'hasArrow', true, 'value', category, 'tooltipTitle', "")
		end

		dewdrop:AddLine()
		dewdrop:AddLine('text', "Weapons", 'hasArrow', true, 'value', "WeaponBuffs", 'tooltipTitle', "")
	elseif (level == 2) then
		for class in pairs(self.buffs) do
			if (value == class.."Buffs") then
				for buff, value in SortTable(self.buffs[class]) do
					if (type(value) == "table") then
						dewdrop:AddLine('text', buff, 'hasArrow', true, 'value', class..buff.."Buffs", 'tooltipTitle', "")
					else
						dewdrop:AddLine('text', buff, 'func', function(buff) self:SetBuff(buff) self:UpdateBuffs() end, 'checked', self:IsChecked(buff), 'arg1', buff, 'tooltipTitle', "")
					end
				end
			end
		end
		for category in pairs(self.others) do
			if (value == category) then
				for buff in SortTable(self.others[category]) do
					dewdrop:AddLine('text', buff, 'func', function(buff) self:SetBuff(buff) self:UpdateBuffs() end, 'checked', self:IsChecked(buff), 'arg1', buff, 'tooltipTitle', "")
				end
			end
		end
		if (value == "WeaponBuffs") then
			dewdrop:AddLine('text', "Main Hand", 'hasArrow', true, 'value', "WeaponMH", 'tooltipTitle', "")
			dewdrop:AddLine('text', "Off Hand", 'hasArrow', true, 'value', "WeaponOH", 'tooltipTitle', "")
		end
	elseif (level == 3) then
		if (value == "PriestDivine SpiritBuffs") then
				dewdrop:AddLine('text', "Group", 'func', function(buff) self:SetBuff("Prayer of Spirit") self:UpdateBuffs() end, 'checked', self:IsChecked("Prayer of Spirit"), 'arg1', buff, 'tooltipTitle', "")
				dewdrop:AddLine('text', "Solo", 'func', function(buff) self:SetBuff("Divine Spirit") self:UpdateBuffs() end, 'checked', self:IsChecked("Divine Spirit"), 'arg1', buff, 'tooltipTitle', "")
				--dewdrop:AddLine('text', "Either", 'tooltipTitle', "")
		elseif (value == "PriestPower Word: FortitudeBuffs") then
				dewdrop:AddLine('text', "Group", 'func', function(buff) self:SetBuff("Prayer of Fortitude") self:UpdateBuffs() end, 'checked', self:IsChecked("Prayer of Fortitude"), 'arg1', buff, 'tooltipTitle', "")
				dewdrop:AddLine('text', "Solo", 'func', function(buff) self:SetBuff("Power Word: Fortitude") self:UpdateBuffs() end, 'checked', self:IsChecked("Power Word: Fortitude"), 'arg1', buff, 'tooltipTitle', "")
				--dewdrop:AddLine('text', "Either", 'tooltipTitle', "")
		elseif (value == "PriestShadow ProtectionBuffs") then
				dewdrop:AddLine('text', "Group", 'func', function(buff) self:SetBuff("Prayer of Shadow Protection") self:UpdateBuffs() end, 'checked', self:IsChecked("Prayer of Shadow Protection"), 'arg1', buff, 'tooltipTitle', "")
				dewdrop:AddLine('text', "Solo", 'func', function(buff) self:SetBuff("Shadow Protection") self:UpdateBuffs() end, 'checked', self:IsChecked("Shadow Protection"), 'arg1', buff, 'tooltipTitle', "")
				--dewdrop:AddLine('text', "Either", 'tooltipTitle', "")
		elseif (value == "DruidMark of the WildBuffs") then
				dewdrop:AddLine('text', "Group", 'func', function(buff) self:SetBuff("Gift of the Wild") self:UpdateBuffs() end, 'checked', self:IsChecked("Gift of the Wild"), 'arg1', buff, 'tooltipTitle', "")
				dewdrop:AddLine('text', "Solo", 'func', function(buff) self:SetBuff("Mark of the Wild") self:UpdateBuffs() end, 'checked', self:IsChecked("Mark of the Wild"), 'arg1', buff, 'tooltipTitle', "")
				--dewdrop:AddLine('text', "Either", 'tooltipTitle', "")
		elseif (value == "MageArcane IntellectBuffs") then
				dewdrop:AddLine('text', "Group", 'func', function(buff) self:SetBuff("Arcane Brilliance") self:UpdateBuffs() end, 'checked', self:IsChecked("Arcane Brilliance"), 'arg1', buff, 'tooltipTitle', "")
				dewdrop:AddLine('text', "Solo", 'func', function(buff) self:SetBuff("Arcane Intellect") self:UpdateBuffs() end, 'checked', self:IsChecked("Arcane Intellect"), 'arg1', buff, 'tooltipTitle', "")
				--dewdrop:AddLine('text', "Either", 'tooltipTitle', "")
		elseif (value == "WeaponMH") then
			for buff in SortTable(self.weapons) do
				dewdrop:AddLine('text', buff, 'func', function(buff) self:SetTempBuff(buff, 1) self:UpdateBuffs() end, 'arg1', buff, 'checked', self:IsTempChecked(buff, 1), 'tooltipTitle', "")
			end
		elseif (value == "WeaponOH") then
			for buff in SortTable(self.weapons) do
				dewdrop:AddLine('text', buff, 'func', function(buff) self:SetTempBuff(buff, 2) self:UpdateBuffs() end, 'arg1', buff, 'checked', self:IsTempChecked(buff, 2), 'tooltipTitle', "")
			end
		end
	end
end

function BuffMe:MenuSettings2(level, value)
	if (level == 1) then
		dewdrop:AddLine('text', L["BuffMe! Options"], 'isTitle', true)
		dewdrop:AddLine()
		for class in SortTable(self.buffs) do
			dewdrop:AddLine('text', class, 'hasArrow', true, 'value', class.."Buffs", 'tooltipTitle', "")
		end
		dewdrop:AddLine()
		for category in pairs(self.others) do
			dewdrop:AddLine('text', category, 'hasArrow', true, 'value', category, 'tooltipTitle', "")
		end
		dewdrop:AddLine()
		dewdrop:AddLine('text', "Weapons", 'hasArrow', true, 'value', "WeaponBuffs", 'tooltipTitle', "")
	elseif (level == 2) then
	elseif (level == 3) then
	end
end

function BuffMe:UpdateBuffs()
	local opt = self.db.profile
	local numBuffs = 0

	for buff, state in pairs(self:GetActiveBuffs()) do
		if (opt.hide) and (state) then
		else
			numBuffs = numBuffs + 1
			self:SetBuffMeBuff(numBuffs, numBuffs + 1, buff, state)
		end
	end

	for i = 1, 2 do
		if (self.db.char.weaponBuffs[i]) and (self:HasTempBuff(self.db.char.weaponBuffs[i], i)) then
			if (not opt.hide) then
				numBuffs = numBuffs + 1
				self:SetBuffMeBuff(numBuffs, numBuffs + 1, self.db.char.weaponBuffs[i], true, i)
			end
		elseif (self.db.char.weaponBuffs[i]) then
			numBuffs = numBuffs + 1
			self:SetBuffMeBuff(numBuffs, numBuffs + 1, self.db.char.weaponBuffs[i], false, i)
		end
	end

	self:Cleanup(numBuffs)
end

function BuffMe:SetBuffMeBuff(buttonid, id, buffname, bstatus, slot)
	self:CreateButton(buttonid)

	local icon = getglobal("BuffMeFrameBuff"..buttonid.."Icon")
	local buff = getglobal("BuffMeFrameBuff"..buttonid.."Text")

	icon:SetTexture(BS:GetSpellIcon(buffname) or self.workarounds[buffname])

	if (bstatus) then
		statustext = "|cff00ff00"..L["ON"].."|r"
	else
		statustext = "|cffff0000"..L["OFF"].."|r"
	end

	if (slot == 1) then
		buffname = buffname.." (MH)"
	elseif (slot == 2) then
		buffname = buffname.." (OH)"
	end

	buff:SetText(buffname..": "..statustext)

	icon:Show()
	buff:Show()
end

function BuffMe:CreateButton(numButtons)
	if (numButtons > self.numBuffs) then
		for i = self.numBuffs + 1, numButtons do
			local f = getglobal("BuffMeFrame")
			local tx = f:CreateTexture("BuffMeFrameBuff"..i.."Icon", "ARTWORK")
			tx:SetWidth(16)
			tx:SetHeight(16)
			tx:ClearAllPoints()
			tx:SetPoint("TOPLEFT", "BuffMeFrame", "TOPLEFT", 12, -20 * (i - 1) - 12)

			local ft = f:CreateFontString("BuffMeFrameBuff"..i.."Text", "OVERLAY")
			ft:SetFontObject(GameFontHighlightSmall)
			ft:SetJustifyH("LEFT")
			ft:SetWidth(self.db.profile.width - 24)
			ft:SetHeight(12)
			ft:ClearAllPoints()
			ft:SetPoint("TOPLEFT", "BuffMeFrameBuff"..i.."Icon", "TOPRIGHT", 4, 0)

			BuffMeFrame:SetHeight((i + 1) * 20)
		end
		self.numBuffs = numButtons
	end
end

function BuffMe:Cleanup(numBuffs)
	local opt = self.db.profile
	local b = getglobal("BuffMeFrame")

	if (numBuffs < self.numBuffs) then
		for i = numBuffs + 1, self.numBuffs do
			local icon = getglobal("BuffMeFrameBuff"..i.."Icon")
			local buff = getglobal("BuffMeFrameBuff"..i.."Text")
			icon:Hide()
			buff:Hide()
		end
	end

	if (opt.autohide) and (numBuffs < 1) then
		b:Hide()
	else
		b:SetHeight((numBuffs + 1) * 20)
		b:Show()
	end
end

function BuffMe:GetActiveBuffs()
	local opt = self.db.char
	local buffs = compost:Acquire()

	for i, value in pairs(opt.buffs) do
		bstatus = (aura:UnitHasBuff("player", value)) or (aura:UnitHasBuff("player", self.buffs["Shaman"][value]))
		pstatus = (aura:UnitHasBuff("player", value)) or (aura:UnitHasBuff("player", self.others["Consumables"][value]))
		if (bstatus or pstatus) then
			buffs[value] = true
		else
			buffs[value] = false
		end
	end
	return buffs
end

function BuffMe:GetTempBuffs()
	local tempBuffs = compost:Acquire()

	for i = 16, 17 do
		gratuity:SetInventoryItem("player", i)

		local _, _, tmpbuff = gratuity:Find("^([^%(]+) %(%d+ [^%)]+%)")

		if (tmpbuff1) then
			tempBuffs[i - 15] = true
		else
			tempBuffs[i - 15] = false
		end
	end
	return tempBuffs
end

function BuffMe:HasTempBuff(buff, slot)
	gratuity:SetInventoryItem("player", slot + 15)

	local _, _, tmp = gratuity:Find("^([^%(]+) %(%d+ [^%)]+%)")

	if (buff == tmp) then
		return true
	end
	return false
end

function BuffMe:SetBuff(buff)
	local opt = self.db.char

	if (getn(opt.buffs) > 0) then
		for i, value in pairs(opt.buffs) do
			if (value == buff) then
				table.remove(opt.buffs, i)
				return
			end
		end
		table.insert(opt.buffs, buff)
	else
		table.insert(opt.buffs, buff)
	end
end

function BuffMe:IsChecked(buff)
	for i, value in pairs(self.db.char.buffs) do
		if (value == buff) then
			return true
		end
	end
	return false
end

function BuffMe:SetTempBuff(buff, slot)
	local opt = self.db.char

	if (buff == opt.weaponBuffs[slot]) then
		opt.weaponBuffs[slot] = nil
	else
		opt.weaponBuffs[slot] = buff
	end
end

function BuffMe:IsTempChecked(buff, slot)
	local opt = self.db.char

	if (buff == opt.weaponBuffs[slot]) then
		return true
	end
	return false
end